// Listing10-3.S
//
// 32-bit ARM code for the double-precision to 
// string conversion.

#include    "ao32aa.inc"

// Number of iterations for testing:

#define iterations 10000000

// Maximum number of signficant digits for
// a single-precision value:

#define maxFPdigits 16

// The following are required to use ARM V7&8 instructions
// on Raspberry Pi devices (3, 4, & 5):

    .fpu    neon-vfpv4
    .arch   armv8-a

    .data
    
// Output buffer for string conversion:

outputBuf:
    .space  256, 0

// Test values:
    
    .align  2   
fpValues:
    .double 0.0
    .dword  1                   // Smallest denormalized value.
    .dword  0x7fefffffffffffff  // Largest value
    .double 1.234567890123456e-9
    .double 1.0000000e-8
    .double 1.0000000e-7
    .double 1.0000000e-6
    .double 1.0000000e-5
    .double 1.0000000e-4
    .double 1.0000000e-3
    .double 1.0000000e-2
    .double 1.0000000e-1
    .double 1.0000000e0
    .double 1.0000000e1
    .double 1.0000000e2
    .double 1.0000000e3
    .double 1.0000000e4
    .double 1.0000000e5
    .double 1.0000000e6
    .double 1.0000000e7
    .double 1.0000000e8
    .double 1.0000000e9
    .double 1.6777214e0
    .double 1.6777215e0
    .double 1.6777216e0
    .double 9.9999999e0
    .double 1.6777214e7
    .double 1.6777215e7
    .double 1.6777216e7
    .double 9.9999999e7
    .dword  0x7ff0000000000000      // +INF
    .dword  0xfff0000000000000      // -INF
    .dword  0x7fffffffffffffff      // NaN
    .dword  0xffffffffffffffff      // Also NaN
  .set    numFP, (.-fpValues)/8

// Large lookup tables used by FPDigits:

    #include    "r64Tbl.S"
    #include    "r64ExpTbl.S"
    #include    "fourDigitTbl.S"    


//***********************************************************        
//
// Include the print routine.

    .align  2

    #include    "print.inc"        

    
//***********************************************************        
//
// getTitle returns the address of a string in
// the R0 register. The c.cpp program will print
// this string before and after calling asmMain.

title:  
    wastr  "Listing10-3"
        
    proc    getTitle, public
    adr     r0, title
    bx      lr
    endp    getTitle


//***********************************************************
//                                                           
// r64ToStr                                                 
//                                                           
// Converts a double-precision floating-point number to the       
// corresponding string of digits.  Note that this           
// function always emits the string using decimal            
// notation.  For scientific notation, use the e64ToStr      
// routine.                                                  
//                                                           
// On Entry:                                                 
//                                                           
// D0-      Double-precision value to convert.
//                                                           
// R0-      r64ToStr stores the resulting characters in  
//          this buffer.
//
// R1-      Field width for the number (note that this   
//          is an *exact* field width, not a minimum         
//          field width).
//                                                    
// R2-      # of digits to display after the decimal pt.
//                                                    
// R3-      Padding character if the number is smaller   
//          than the specified field width.
//                                                                     
// R4-      Maximum string length.
//                                                                            
// On Exit:                                                  
//                                                           
// Buffer contains the newly formatted string.  If the       
// formatted value does not fit in the width specified,      
// r64ToStr will store "#" characters into this string.      
//                                                           
// Carry-    Clear if success, set if an exception occurs.                                                         
//           If width is larger than the maximum length of          
//           the string specified by buffer, this routine        
//           will return with the carry set.                                             
//                                                           
//***********************************************************

// r64ToStr will preserve R0:

    proc    r64ToStr, thumb, public
    push    {r0, lr}
    bl      r64ToBuf
    bcs     returnError
    pop     {r0, pc}
    
returnError:
    add     sp, sp, #4
    pop     {pc}
    endp    r64ToStr
            
////////////////////////////////////////////////////////////
//
// r64ToBuf will leave R0 pointing at the end of the string:

    .macro  r64Entry
    push    {r1-r11, lr}    // Note: must keep stack
    vpush   {d0-d1}         //       8-byte aligned!
    sub     sp, sp, #40     // Local storage for fp digits
    .endm
    
    .macro  r64Rtn
    add     sp, sp, #40
    vpop    {d0-d1}
    pop     {r1-r11, pc}
    .endm
    
    proc    r64ToBuf, thumb, public
    r64Entry

    // Use meaningful names for the non-volatile
    // registers that hold local/parameter values:
                
    #define exponent    r8
    #define fWidth      r9
    #define decDigits   r10
    #define outBuf      r11
    #define fill        [sp, #34]
    #define sign        [sp, #35]
         
    // First, make sure the number will fit into
    // the  specified string.
    
    cmp     r4, #5          // Must have at least 5 bytes.
    blo     valOutOfRange
    
    cmp     r1, r4          // R1=width, r4=buf size 
    bhs     strOverflow
        
    // Check for NaN, INF, and -INF special cases.
    
    bl      chkNanInf64
    bcc     notSpecial
    adds    r0, r0, #0      // Clear carry, no error.
    r64Rtn
                
    // Move parameter values to non-volatile 
    // storage:

notSpecial:
    strb    r3, fill 
    mov     decDigits,  r2 
    mov     fWidth,     r1 
    mov     outBuf,     r0 
    
    //  If the width is zero, return an error:

    cmp     fWidth, #0
    beq     valOutOfRange
    
    // Okay, do the conversion.  Begin by
    // processing the mantissa digits.
                
    mov     r0, sp              // Buffer for FPdigits        
    bl      FPDigits            // Convert D0 to string.

    mov     exponent, r1        // Save away exponent result.
    strb    r2, sign            // Save mantissa sign char.

// Round the string of digits to the number of significant 
// digits we want to display for this number. Note that there 
// are a maximum of 16 digits produced for a 53-bit value.
//
// First, if the exponent is positive or zero, add 1 because
// FPDigits assumes the decimal point is to the right of the
// most significant digit; don't add one if it's negative
// because the decimal point is to the left of the MSD.

    cmp     r1, #0          // R1=abs(exponent)
    it      ge
    addge   r1, r1, #1

	// If the exponent is negative, or the number of
	// digits is too large, force to zero:
	
    cmp     r1, #maxDPdigits
	it		hi
    movhi   r1, #0          // If the exponent is negative, or
                            // too large, set width to 0.

    add     r1, r1, decDigits // Compute rounding position
    cmp     r1, #maxFPdigits
    bhs     dontRound         // Don't bother if a big #.

    // To round the value to the number of
    // significant digits, go to the digit just
    // beyond the last one we are considering (R1
    // currently contains this position) and add 5 
	// to that digit. Propagate any overflow into 
	// the remaining digit positions.
            
    ldrb    r0, [sp, r1]        // Get that digit.

    add     r0, r0, #5          // Round (e.g., +0.5 ).
    cmp     r0, #'9'
    bls     dontRound

    mov     r0, #('0' + 10)     // Force to zero
whileDigitGT9:                          
    sub     r0, r0, #10         // Sub out overflow,
    strb    r0, [sp, r1]        // carry, into prev

    subs    r1, r1, #1          // digit (until 1st
    bmi     hitFirstDigit       //  digit in the #)
     
    ldrb    r0, [sp, r1]        // Increment previous
    add     r0, r0, #1          // digit.
    strb    r0, [sp, r1]                           

    cmp     r0, #'9'            // Overflow if > '9'
    bhi     whileDigitGT9
    b       dontRound
 
hitFirstDigit:

    // If we get to this point, then we've hit the
    // first digit in the number.  So we've got to
    // shift all the characters down one position
    // in the string of bytes and put a "1" in the
    // first character position.
    
    mov     r2, #maxFPdigits      // Max digits in value
repeatUntilX2eq0:
    ldrb    r0, [sp, r2]
    add     r2, r2, #1
    strb    r0, [sp, r2]
    subs    r2, r2, #2
    bne     repeatUntilX2eq0
    
    mov     r0, #'1'
    strb    r0, [sp, r2]
     
    add     exponent, exponent, #1 // Increment exponent because
                                   // we added a digit
dontRound:
    // Handle positive and negative exponents separately.

    mov     r5, #0             // Index into output buf
    cmp     exponent, #0
    bge     positiveExponent

    // Negative exponents:
    // Handle values between 0 & 1.0 here (negative
    // exponents imply negative powers of ten).
    // 
    // Compute the number's width.  Since this
    // value is between 0 & 1, the width
    // calculation is easy: it's just the number of
    // decimal positions they've specified plus
    // three (since we need to allow room for a
    // leading "-0."). R2=number of digits to emit
    // after "."

    mov     r4, #4            
    add     r2, decDigits, #3
    cmp     r2, r4
    it      lo
    movlo   r2, r4      // If X2<X4, X2=X4

    cmp     r2, fWidth
    bhi     widthTooBig 
    
    // This number will fit in the specified field
    // width, so output any necessary leading pad
    // characters. R3=number of padding characters
    // to output.
    
    sub     r3, fWidth, r2
    b       testWhileX3ltWidth
        
whileX3ltWidth:
    ldrb    r0, fill
    strb    r0, [outBuf, r5]
    add     r5, r5, #1          // Index
    add     r2, r2, #1          // Digits processed
testWhileX3ltWidth:
    cmp     r2, fWidth
    blo     whileX3ltWidth
                    
    // Output " 0." or "-0.", depending on 
    // the sign of the number.
    
    ldrb    r0, sign
    strb    r0, [outBuf, r5]
    add     r5, r5, #1
    mov     r0, #'0'
    strb    r0, [outBuf, r5]
    add     r5, r5, #1
    mov     r0, #'.'            
    strb    r0, [outBuf, r5]
    add     r5, r5, #1
    add     r3, r3, #3        
    
    // Now output the digits after the decimal point:

    mov     r2, #0              // Count the digits here
    mov     r1, sp              // Point at local storage
                                    
// If the exponent is currently negative, or if
// we've output more than 16 significant digits,
// just output a zero character.
            
repeatUntilX3geWidth:
    mov     r0, #'0'
    adds    exponent, exponent, #1
    bmi     noMoreOutput
    
    cmp     r2, #maxFPdigits
    bge     noMoreOutput
    
    ldrb    r0, [r1], #1

noMoreOutput:
    strb    r0, [outBuf, r5]
    add     r5, r5, #1          // Index
    add     r2, r2, #1          // Digits processed
    add     r3, r3, #1          // Digit count
    cmp     r3, fWidth
    blo     repeatUntilX3geWidth
    b       r64BufDone

// If the number's actual width was bigger than the width
// specified by the caller, emit a sequence of '#' characters
// to denote the error.
                            
widthTooBig:    

    // The number won't fit in the specified field
    // width, so fill the string with the "#"
    // character to indicate an error.
            
    mov     r2, fWidth
    mov     r0, #'#'
fillPound:  
    strb    r0, [outBuf, r5]
    add     r5, r5, #1          // Index
    subs    r2, r2, #1
    bne     fillPound
    b       r64BufDone
            
// Handle numbers with a positive exponent here.
//
// Compute # of print positions consumed by output string.
// This is given by:
//
//           Exponent     // # of digits to left of "."
//   +       2            // Sign + 1's digit
//   +       decDigits    // Add in digits right of "."
//   +       1            // If there is a decimal point
            
positiveExponent:
    mov     r3, exponent    // Digits to left of "."
    add     r3, r3, #2      // sign posn.
    cmp     decDigits, #0   // See if any fractional
    beq     decPtsIs0       // part

    add     r3, r3, decDigits // Digits to right of "."
    add     r3, r3, #1        // Make room for the "."
    
    // Make sure the result will fit in the
    // specified field width.
            
decPtsIs0:
    cmp     r3, fWidth
    bhi     widthTooBig
    beq     noFillChars
            
    // If the actual number of print positions
    // is less than the specified field width,
    // output leading pad characters here.
    
    subs    r2, fWidth, r3
    beq     noFillChars
    
    ldrb    r0, fill
fillChars:  
    strb    r0, [outBuf, r5]
    add     r5, r5, #1
    subs    r2, r2, #1
    bne     fillChars
                    
noFillChars:
 
    // Output the sign character.
    
    ldrb    r2, sign
    strb    r2, [outBuf, r5]
    add     r5, r5, #1
            
    // Okay, output the digits for the number here.
    
    mov     r2, #0              // Counts  # of output chars.
    mov     r1, sp              // Point at local storage            
    
    // Calculate the number of digits to output
    // before and after the decimal point.
    
    add     r3, decDigits, exponent
    add     r3, r3, #1          // Always one digit before "."
                    
// If we've output fewer than 16 digits, go ahead
// and output the next digit.  Beyond 16 digits,
// output zeros.
            
repeatUntilX3eq0:
    mov     r0, #'0'
    cmp     r2, #maxFPdigits
    bhs     putChar
    
    ldrb    r0, [r1], #1
putChar:
    strb    r0, [outBuf, r5]
    add     r5, r5, #1     
    
    // If the exponent decrements down to zero,
    // then output a decimal point.
    
    cmp     exponent, #0
    bne     noDecimalPt
    
    cmp     decDigits, #0
    beq     noDecimalPt
    
    mov     r0, #'.'
    strb    r0, [outBuf, r5]
    add     r5, r5, #1
                    
noDecimalPt:
    sub     exponent, exponent, #1  // Count down to "." output
    add     r2, r2, #1              // # of digits thus far
    subs    r3, r3, #1              // Total # of digits to output
    bne     repeatUntilX3eq0

// Zero-terminate string and leave:
            
r64BufDone: 
    mov     lr, #0
    strb    lr, [outBuf, r5]
    
    // Note that the following instruction will always
    // clear the carry flag, denoting no error (adding
    // an offset and an address will not produce an
    // overflow).
    
    adds    r0, outBuf, r5  // Point R0 at end of string
    r64Rtn

// Drop down here if there was an error creating the string:

strOverflow:
    mov     r0, #-3 //String overflow
    b       ErrorExit

valOutOfRange:       
    mov     r0, #-1         // Range error
ErrorExit:  
    cmp     r0, r0          // stc=error
            
    // Change r0 on return:

    r64Rtn
    endp    r64ToBuf                   

///////////////////////////////////
//
// chkNanInf64
//
// Checks D0 to see if it is NaN,
// INF, or -INF.
//
// On entry:
//
//  D0- Value to test
//  R1- Field width
//
// Note: modifies and does not preserve R5, R6, and R7.

    proc    chkNanInf64, thumb, public
    
    // Handle NaN and INF special cases:
    
    ldr     r7, =0x7ff00000
    vmov    r5, r6, D0
    and     r6, r6, r7
    cmp     r6, r7
    beq     isSpecial
    
    // Not NaN or INF, so return with the carry clear:
    
    bx      lr

// The exponent is 0x7FF (NaN or INF). It's INF if
// the mantissa is zero, NaN otherwise
    
isSpecial:
    vmov    r5, r6, d0
    cmp     r5, #0
    bne     isNaN
    
    ldr     r7, =0x000fffff
    tst     r7, r6
    beq     isINF
    
    // If the mantissa is non-zero, then
    // it must be NaN:
    
isNaN:
    ldr     r7, ='N' + ('a' << 8) + ('N' << 16)
    str     r7, [r0]
    mov     r7, #3
    b       fillSpecial
    
// The sign bit determines whether this is -INF
// or +INF:
    
isINF:
    ldr     r7, =0x80000000
    tst     r7, r6
    bne     minusINF

plusINF:    
    ldr     r7, ='I' + ('N' << 8) + ('F' << 16)
    str     r7, [r0]
    mov     r7, #3
    b       fillSpecial
    
minusINF:
    ldr     r7, ='-' + ('I'<<8) + ('N'<<16) + ('F'<<24)
    str     r7, [r0]
    mov     r7, #4

// For NaN and INF, fill the remainder of the string, as appropriate:
    
fillSpecial:
    mov     r6, #' '
    b       whlLTwidth
    
fsLoop:     
    strb    r6, [r0, r7]
    add     r7, r7, #1
whlLTwidth:
    cmp     r7, r1
    blo     fsLoop 
    cmp     r7, r7 // Set carry to indicate special value

    bx      lr
    endp    chkNanInf64
                        
/////////////////////////////////////////////////////////
//
// FPDigits:
//
//  Used to convert a floating-point value
//  in D0 to a string of digits.
//
// Inputs:
//      D0: Double-precision value to convert
//      R0: Pointer to buffer to receive chars
//
// Outputs:
//      R0: Still points at buffer
//      R1: Contains exponent of the number
//      R2: Contains sign (space or '-')
 
    proc        FPDigits, thumb, public

    push        {r3, r4, r5, r6, r7, r9, r10, lr}
    vpush       {d0, d1, d2, d3}
            
    // Save sign in R9 for the time being:
    
    mov         r9, #' '        // Assume sign is +

    // Special case for 0.0:
    
    vcmp.f64    d0, #0.0
    vmrs        APSR_nzcv, FPSCR
    bne         d0not0
    ldr         r1, =0x30303030
    str         r1, [r0]
    str         r1, [r0, #4]
    str         r1, [r0, #8]
    str         r1, [r0, #12]
    mov         r1, #0          // Exponent = 0
    strb        r1, [r0, #16]   // Also zero terminator
    mov         r2, #' '        // Sign is positive
    b           fpdDone
            
// If the number is non-zero, deal with it here. Note
// that the flags were set by comparing D0 to 0.0 earlier.

d0not0:     
    bge      fpIsPositive    // See if positive or negative
        
    // If negative, negate and change the sign
    // character to '-'
    
    vabs.f64    d0, d0
    mov         r9, #'-'

// Get the number between 1.0 and <10.0 so we can figure out 
// what the exponent is.  Begin by checking to see if we have 
// a positive or negative exponent.

fpIsPositive:

    mov         r1, #0          // Initialize exponent
    ldr         r6, =0x7ff00000 // Mask out exponent
    vldr.f64    d1, ten

    // Handle denormalized numbers:

denormLoop: 
    vmov        r3, r7, D0
    ands        r7, r6
    bne         notDenorm
    
    // Multiply value by 10.0:
    
    vmul.f64    D0, D0, D1
    sub         r1, r1, #1      // Decrement exponent
    b           denormLoop
        
// Use the exponent as an index to the r64Tbl to determine
// how to get this value in the range 1.0 <= D0 < 10.0:

notDenorm:
    vmov        r6, r7, d0
    ldr         r6, =0x7ff
    lsr         r7, r7, #20
    and         r7, r7, r6
    
    ldr         r5, =r64ExpTbl
    lsl         r7, r7, #2      // *4 is index into r64ExpTbl.
    ldr         r4, [r5, r7]    // Get integer exponent.
    lsl         r7, r7, #1      // *8 is index into r64Tbl.
    ldr         r6, =r64Tbl
    add         r7, r6, r7
    vldr        d1, [r7]        // Get multiplier value.
    add         r1, r1, r4      // Adjust exponent per table.
    vmul.f64    d0, d1

// Okay, at this point the number is probably in the range 1 <= x < 10,
// but for certain values, it may be slightly less than 1.0. If so,
// multiply by 10 and subtract one from the exponent:

    vmov.f64    d1, #1.0
    vcmp.f64    d0, d1
    vmrs        APSR_nzcv, FPSCR
    bhs         inRange1_10
    
    vmov.f64    d1, #10.0
    vmul.f64    d0, d1
    sub         r1, r1, #1
    
// Okay, the value is definitely in the range 1.0 <= DO < 10.0.
// Process that value here:
            
inRange1_10:
    mov           r10, r1         // Preserve exponent
    ldr           r1, =fourDigitTbl
    
    // Convert HO four digits (12-15) to an integer:
    
    vldr.f64      d1, e3
    vmul.f64      d2, d0, d1    // Value in range 1 to 9999.
    vcvt.s32.f64  s0, d2        // Convert to integer and
    vmov          r2, s0        // use as index into fourDigitTbl. 
    ldr           r3, [r1, r2, lsl #2]
    str           r3, [r0]
    
    // Convert digits 8-11 to an integer:
    
    vcvt.f64.s32  d3, s0
    vsub.f64      d3, d2, d3
    vldr.f64      d1, e4
    vmul.f64      d3, d3, d1
    vcvt.s32.f64  s0, d3
    vmov          r2, s0        // use as index into fourDigitTbl. 
    ldr           r3, [r1, r2, lsl #2]
    str           r3, [r0, #4]
    
    // Convert digits 4-7 to an integer:
    
    vcvt.f64.s32  d2, s0
    vsub.f64      d2, d3, d2
    vldr.f64      d1, e4
    vmul.f64      d2, d2, d1
    vcvt.s32.f64  s0, d2
    vmov          r2, s0        // use as index into fourDigitTbl. 
    ldr           r3, [r1, r2, lsl #2]
    str           r3, [r0, #8]
    
    // Convert digits 0-3 to an integer:
    
    vcvt.f64.s32  d3, s0
    vsub.f64      d3, d2, d3
    vldr.f64      d1, e4
    vmul.f64      d3, d3, d1
    vcvtr.s32.f64 s0, d3
    vmov          r2, s0        // use as index into fourDigitTbl. 
    ldr           r3, [r1, r2, lsl #2]
    str           r3, [r0, #12]
                                  
    mov           r2, r9  // Return sign in R2.
    mov           r1, r10 // Return exponent in R1.
     
fpdDone:   
    vpop          {d0, d1, d2, d3}
    pop           {r3, r4, r5, r6, r7, r9, r10, pc}
    endp          FPDigits

// Floating-point constants used by the FPDigits code:

    .align      3
e4:
    .double     1e4
    
e3:
    .double     1e3

ten:        
    .double     10.0
    
/////////////////////////////////////////////////////////////////////
//              
// Here's the assembly language main program:
 
    proc    asmMain, public
    push    {r0-r12, lr}
        
///////////////
// Timing code:

    ldr     r10, =iterations
timingLp:

    ldr     r9, =fpValues
    mov     r8, #numFP
innerLp:
    ldr     r0, =outputBuf     
    vldr    d0, [r9]
    add     r9, r9, #8
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1:
    
    mov     r1, #24         // Width
    mov     r2, #8          // Dec posns
    mov     r3, #' '        // Pad char
    mov     r4, #256        // Max buffer
    bl      r64ToBuf

    // Repeat for each sample input:
      
    subs    r8, #1
    bne     innerLp            

    // Repeat a large number of times in order
    // to time the code:
    
    subs    r10, r10, #1
    bne     timingLp

///////////////////////////
// Output code for testing:

    ldr     r9, =fpValues
    mov     r8, #numFP
convLp:
    ldr     r0, =outputBuf     
    vldr    d0, [r9]
       
    // Convert string pointed at by R0 to 
    // a numeric value in R1 and output
    // that value:
    
    mov     r1, #23         // Width
    mov     r2, #9          // Dec posns
    mov     r3, #' '        // Pad char
    mov     r4, #256        // Max buffer
    bl      r64ToBuf
    bcc     goodConv
    
    bl      print
    wastr   "Error %ld converting string\n"
    b       nextConv

goodConv:
    vldr    d0, [r9]
    add     r9, r9, #8
    vmov    r1, r2, d0
    bl      print
    wastr   "Converted %20.8e to "
    ldr     r0, =outputBuf
    bl      print
    wastr   "'%s'\n"
 
          
nextConv:
    subs    r8, #1
    bne     convLp
    pop     {r0-r12, pc} 
    endp    asmMain


